import React, { useContext } from "react";
import {Row, Table} from "react-bootstrap";
import { Chart as ChartJS, CategoryScale, LinearScale, BarElement, Title, Tooltip, Legend } from 'chart.js';

import { ChargingStructure } from "../Schema";
import { createChargingStructureDataLookup } from "../helpers/dataconversion";
import ColorPill from "../components/ColorPill";
import DataPage from "../components/DataPage";
import Filter from "../components/graphing/Filter";
import {ExportType, Sections} from "../helpers/constants";
import DownloadDataButton from "../components/DownloadDataButton";
import { FilterSelectionContext } from "../helpers/FilterSelectionProvider";
import DownloadImageChartButton from "../components/DownloadImageChartButton";
import ChartContainer from "../components/graphing/ChartContainer";
import { useData } from "../helpers/useData";

function ChargingStructurePage(): React.ReactElement {
    const {filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const {data: chargingStructureData, years, nrens} = useData<ChargingStructure>('/api/charging/', setFilterSelection);

    const selectedData = (chargingStructureData).filter(data =>
        filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren)
    );

    const dataLookup = createChargingStructureDataLookup(selectedData);

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
        coloredYears
    />

    return (
        <DataPage title="Charging Mechanism of NRENs per Year"
        description="The charging structure is the way in which NRENs charge their customers for the services they provide.
         The charging structure can be based on a flat fee, usage based fee, a combination of both, or no direct charge. 
         By selecting multiple years and NRENs, the table can be used to compare the charging structure of NRENs."
        category={Sections.Organisation} filter={filterNode}>
            <>
                <Row>
                    <DownloadDataButton data={chargingStructureData} filename="charging_mechanism_of_nrens_per_year.csv" exportType={ExportType.CSV}/>
                    <DownloadDataButton data={chargingStructureData} filename="charging_mechanism_of_nrens_per_year.xlsx" exportType={ExportType.EXCEL} />
                </Row>
                <DownloadImageChartButton filename="charging_mechanism_of_nrens_per_year"/>
                <ChartContainer>
                    <Table className="charging-struct-table" striped bordered responsive>
                        <colgroup>
                            <col span={1} style={{width: "10%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                        </colgroup>
                        <thead>
                        <tr>
                            <th></th>
                            <th>Flat fee based on bandwidth</th>
                            <th>Usage based fee</th>
                            <th>Combination flat fee & usage basedfee</th>
                            <th>No Direct Charge</th>
                            <th>Other</th>
                        </tr>
                        </thead>
                        <tbody>
                        {Array.from(dataLookup.entries()).map(([nren, nrenMap]) => (
                            <tr key={nren}>
                                <td>{nren}</td>
                                {["flat_fee", "usage_based_fee", "combination", "no_charge", "other"].map(discriminator => (
                                    <td key={discriminator}>
                                        {Array.from(nrenMap.entries()).map(([year, chargingMechanism]) => (
                                            <ColorPill key={year} year={year}
                                                       active={chargingMechanism == discriminator}/>
                                        ))}
                                    </td>
                                ))}
                            </tr>
                        ))}
                        </tbody>
                    </Table>
                </ChartContainer>
            </>
        </DataPage>
    );
}
export default ChargingStructurePage;
