import React, { useContext } from "react";
import { Row, Table } from "react-bootstrap";

import { Service, ServiceCategory } from "../Schema";
import { createServiceDataLookup } from "../helpers/dataconversion";
import ColorBadgeService from "../components/ColorBadgeService";
import DataPage from "../components/DataPage";
import Filter from "../components/graphing/Filter";
import { ExportType, Sections } from "../helpers/constants";
import DownloadDataButton from "../components/DownloadDataButton";
import { FilterSelectionContext } from "../helpers/FilterSelectionProvider";
import DownloadImageChartButton from "../components/DownloadImageChartButton";
import ChartContainer from "../components/graphing/ChartContainer";
import { useData } from "../helpers/useData";

const CategoryFriendlyNames = {};
CategoryFriendlyNames[ServiceCategory.network_services] = "network";
CategoryFriendlyNames[ServiceCategory.isp_support] = "isp support";
CategoryFriendlyNames[ServiceCategory.security] = "security";
CategoryFriendlyNames[ServiceCategory.identity] = "identity";
CategoryFriendlyNames[ServiceCategory.collaboration] = "collaboration";
CategoryFriendlyNames[ServiceCategory.multimedia] = "multimedia";
CategoryFriendlyNames[ServiceCategory.storage_and_hosting] = "storage and hosting";
CategoryFriendlyNames[ServiceCategory.professional_services] = "professional";


interface inputProps {
    category: ServiceCategory
}

function ServicesPage({ category }: inputProps): React.ReactElement {
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data: serviceData, years, nrens } = useData<Service>('/api/nren-services/', setFilterSelection);

    const selectedData = (serviceData).filter(data =>
        filterSelection.selectedYears.includes(data.year)
        && filterSelection.selectedNrens.includes(data.nren)
        && data.service_category == category
    );

    const { servicesList, dataLookup } = createServiceDataLookup(selectedData);

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
        coloredYears
    />

    return (
        <DataPage title={"NREN " + CategoryFriendlyNames[category] + " services matrix"}
            description="Hover over the elements for additional information."
            category={Sections.Services} filter={filterNode}>
            <>

                <DownloadDataButton data={serviceData} filename="nren_services.csv" exportType={ExportType.CSV} />
                <DownloadDataButton data={serviceData} filename="nren_services.xlsx" exportType={ExportType.EXCEL} />
                <DownloadImageChartButton filename="nren_services" />
                <ChartContainer>
                    <Table className="charging-struct-table" striped bordered>
                        <colgroup>
                            <col span={1} style={{ width: "10%" }} />
                            {/* <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/>
                            <col span={1} style={{width: "18%"}}/> */}
                        </colgroup>
                        <thead>
                            <tr>
                                <th></th>
                                {servicesList.map(([serviceName, serviceDescription]) => (
                                    <th key={serviceName} data-description={serviceDescription} className="bottom-tooltip">{serviceName}</th>
                                ))}
                            </tr>
                        </thead>
                        <tbody>
                            {Array.from(dataLookup.entries()).map(([nren, nrenServiceMap]) => (
                                <tr key={nren}>
                                    <td>{nren}</td>
                                    {servicesList.map(([serviceName, _]) => (
                                        <td key={serviceName}>
                                            {nrenServiceMap.has(serviceName) &&
                                                filterSelection.selectedYears.map(year => {
                                                    const serviceYears = nrenServiceMap.get(serviceName)!;
                                                    const serviceInfo = serviceYears.get(year);
                                                    return <ColorBadgeService key={year} year={year} active={serviceYears.has(year)} serviceInfo={serviceInfo} />;
                                                })
                                            }
                                        </td>
                                    ))}
                                </tr>
                            ))}
                        </tbody>
                    </Table>
                </ChartContainer>
            </>
        </DataPage>
    );
}
export default ServicesPage;
