"""
This module handles adding responses from the latest 2023 survey into the 2023 presentation_models.
The process is straightforward as the data is new and can be directly inserted into the presentation_models.

For subsequent years like 2024, handling should be almost the same as 2023, only changes in questions
will need to be addressed.

Usage:
    Used in publish_survey API in compendium_v2/routes/survey.py
"""

from decimal import Decimal
from typing import List

from sqlalchemy import delete, select

from compendium_v2.db import db
from compendium_v2.db.presentation_models import BudgetEntry, ChargingStructure, ECProject, ExternalConnections, \
    InstitutionURLs, NrenStaff, ParentOrganization, Policy, SubOrganization, TrafficVolume, ExternalConnection, \
    FundingSource, CentralProcurement, ServiceManagement, ServiceUserTypes, EOSCListings, \
    Standards, CrisisExercises, SecurityControls, ConnectedProportion, ConnectivityLevel, \
    ConnectionCarrier, ConnectivityLoad, ConnectivityGrowth, CommercialConnectivity, \
    CommercialChargingLevel, RemoteCampuses, DarkFibreLease, DarkFibreInstalled, FibreLight, \
    NetworkMapUrls, MonitoringTools, PassiveMonitoring, TrafficStatistics, SiemVendors, \
    CertificateProviders, WeatherMap, PertTeam, AlienWave, Capacity, NonREPeers, TrafficRatio, \
    OpsAutomation, NetworkFunctionVirtualisation, NetworkAutomation, NRENService
from compendium_v2.db.presentation_model_enums import CarryMechanism, CommercialCharges, YesNoPlanned, \
    CommercialConnectivityCoverage, ConnectivityCoverage, FeeType, MonitoringMethod, ServiceCategory, UserCategory
from compendium_v2.db.survey_models import ResponseStatus, SurveyResponse


def int_or_none(answers_dict, key):
    if key in answers_dict:
        return int(answers_dict[key])
    return None


def decimal_or_none(answers_dict, key):
    if key in answers_dict:
        return Decimal(answers_dict[key])
    return None


def decimal_or_zero(answers_dict, key):
    return Decimal(answers_dict.get(key, 0))


def bool_or_none(answer, key=None):
    if key:
        answer = answer.get(key)
    if answer:
        return answer == "Yes"
    return None


def _map_2023(nren, answers) -> None:
    year = 2023

    for table_class in [BudgetEntry, ChargingStructure, ECProject, ExternalConnections,
                        InstitutionURLs, NrenStaff, ParentOrganization, Policy, SubOrganization, TrafficVolume,
                        FundingSource, CentralProcurement, ServiceManagement, ServiceUserTypes, EOSCListings,
                        Standards, CrisisExercises, SecurityControls, ConnectedProportion, ConnectivityLevel,
                        ConnectionCarrier, ConnectivityLoad, ConnectivityGrowth, CommercialConnectivity,
                        CommercialChargingLevel, RemoteCampuses, DarkFibreLease, DarkFibreInstalled, FibreLight,
                        NetworkMapUrls, MonitoringTools, PassiveMonitoring, TrafficStatistics, SiemVendors,
                        CertificateProviders, WeatherMap, PertTeam, AlienWave, Capacity, NonREPeers, TrafficRatio,
                        OpsAutomation, NetworkFunctionVirtualisation, NetworkAutomation, NRENService]:
        db.session.execute(delete(table_class).where(table_class.year == year))  # type: ignore

    answers = answers["data"]
    budget = answers.get("budget")
    if budget:
        db.session.add(BudgetEntry(nren_id=nren.id, nren=nren, year=year, budget=Decimal(budget)))

    funding_source = answers.get("income_sources")
    if funding_source:
        db.session.add(FundingSource(
            nren_id=nren.id, nren=nren, year=year,
            client_institutions=decimal_or_zero(funding_source, "client_institutions"),
            european_funding=decimal_or_zero(funding_source, "european_funding"),
            gov_public_bodies=decimal_or_zero(funding_source, "gov_public_bodies"),
            commercial=decimal_or_zero(funding_source, "commercial"),
            other=decimal_or_zero(funding_source, "other")
        ))

    charging = answers.get("charging_mechanism")
    if charging:
        db.session.add(ChargingStructure(nren_id=nren.id, nren=nren, year=year, fee_type=FeeType[charging]))

    staff_roles = answers.get("staff_roles", {})
    staff_employment_type = answers.get("staff_employment_type", {})
    if staff_roles or staff_employment_type:
        db.session.add(NrenStaff(
            nren_id=nren.id, nren=nren, year=year,
            permanent_fte=decimal_or_zero(staff_employment_type, "permanent_fte"),
            subcontracted_fte=decimal_or_zero(staff_employment_type, "subcontracted_fte"),
            technical_fte=decimal_or_zero(staff_roles, "technical_fte"),
            non_technical_fte=decimal_or_zero(staff_roles, "nontechnical_fte")
        ))

    parent = answers.get("parent_organization_name")
    if parent:
        db.session.add(ParentOrganization(nren_id=nren.id, nren=nren, year=year, organization=parent))

    subs = answers.get("suborganization_details")
    if subs:
        for sub in subs:
            role = sub.get("suborganization_role", "")
            if role == "other":
                role = sub.get("suborganization_role-Comment", "")
            db.session.add(SubOrganization(
                nren_id=nren.id, nren=nren, year=year,
                organization=sub.get("suborganization_name"),
                role=role
            ))

    ec_projects = answers.get("ec_project_names")
    if ec_projects:
        for ec_project in ec_projects:
            if ec_project:
                db.session.add(
                    ECProject(nren_id=nren.id, nren=nren, year=year, project=ec_project.get("ec_project_name"))
                )

    strategy = answers.get("corporate_strategy_url", "")
    policies = answers.get("policies", {})
    if strategy or policies:
        db.session.add(Policy(
            nren_id=nren.id, nren=nren, year=year,
            strategic_plan=strategy,
            environmental=policies.get("environmental_policy", {}).get("url", ""),
            equal_opportunity=policies.get("equal_opportunity_policy", {}).get("url", ""),
            connectivity=policies.get("connectivity_policy", {}).get("url", ""),
            acceptable_use=policies.get("acceptable_use_policy", {}).get("url", ""),
            privacy_notice=policies.get("privacy_notice", {}).get("url", ""),
            data_protection=policies.get("data_protection_contact", {}).get("url", ""),
            gender_equality=policies.get("gender_equality_policy", {}).get("url", "")
        ))

    traffic_estimate = answers.get("traffic_estimate")
    if traffic_estimate:
        db.session.add(TrafficVolume(
            nren_id=nren.id, nren=nren, year=year,
            to_customers=decimal_or_zero(traffic_estimate, "to_customers"),
            from_customers=decimal_or_zero(traffic_estimate, "from_customers"),
            to_external=decimal_or_zero(traffic_estimate, "to_external"),
            from_external=decimal_or_zero(traffic_estimate, "from_external")
        ))

    institution_urls = answers.get("connected_sites_lists")
    if institution_urls:
        urls = [i.get("connected_sites_url", "") for i in institution_urls if i.get("connected_sites_url", "") != ""]
        if urls:
            db.session.add(InstitutionURLs(
                nren_id=nren.id, nren=nren, year=year,
                urls=urls
            ))

    central_procurement = answers.get("central_software_procurement")
    if central_procurement:
        central_procurement = central_procurement == "Yes"
        db.session.add(CentralProcurement(
            nren_id=nren.id, nren=nren, year=year,
            central_procurement=central_procurement,
            amount=decimal_or_none(answers, "central_procurement_amount")
        ))

    formal_service_management_framework = answers.get("formal_service_management_framework")
    service_level_targets = answers.get("service_level_targets")
    if formal_service_management_framework or service_level_targets:
        db.session.add(ServiceManagement(
            nren_id=nren.id, nren=nren, year=year,
            service_management_framework=bool_or_none(formal_service_management_framework),
            service_level_targets=bool_or_none(service_level_targets)
        ))

    service_type_matrix = answers.get("service_matrix", {})
    for user_type, service_types in service_type_matrix.items():
        user_type = UserCategory[user_type]
        service_types = service_types["service_types"]
        for service_type in service_types:
            service_type = ServiceCategory[service_type]
            db.session.add(ServiceUserTypes(
                nren_id=nren.id, nren=nren, year=year,
                user_category=user_type,
                service_category=service_type
            ))

    services_on_eosc_portal_list = answers.get("services_on_eosc_portal_list", [])
    eosc_list = [i.get("service_name") for i in services_on_eosc_portal_list]
    eosc_list = [i for i in eosc_list if i]
    if eosc_list:
        db.session.add(EOSCListings(
            nren_id=nren.id, nren=nren, year=year,
            service_names=eosc_list
        ))

    audits = answers.get("audits")
    business_continuity_plans = answers.get("business_continuity_plans")
    crisis_management_procedure = answers.get("crisis_management_procedure")
    if audits or business_continuity_plans or crisis_management_procedure:
        db.session.add(Standards(
            nren_id=nren.id, nren=nren, year=year,
            audits=bool_or_none(audits),
            audit_specifics=answers.get("audit_specifics", ""),
            business_continuity_plans=bool_or_none(business_continuity_plans),
            business_continuity_plans_specifics=answers.get("business_continuity_plans_specifics", ""),
            crisis_management_procedure=bool_or_none(crisis_management_procedure)
        ))

    crisis_exercises = answers.get("crisis_exercises")
    if crisis_exercises:
        db.session.add(CrisisExercises(
            nren_id=nren.id, nren=nren, year=year,
            exercise_descriptions=crisis_exercises
        ))

    security_controls = answers.get("security_controls")
    if security_controls:
        if "other" in security_controls:
            security_controls.remove("other")
            security_controls.append(answers.get("security_controls-Comment", "other"))
        db.session.add(SecurityControls(
            nren_id=nren.id, nren=nren, year=year,
            security_control_descriptions=security_controls
        ))

    connectivity_proportions = answers.get("connectivity_proportions", {})
    for user_type, connectivity_proportion in connectivity_proportions.items():
        user_type = UserCategory[user_type]
        coverage = connectivity_proportion.get("covered")
        coverage = ConnectivityCoverage[coverage] if coverage else None
        number_connected = int_or_none(connectivity_proportion, "nr_connected")
        market_share = decimal_or_none(connectivity_proportion, "market_share_percentage")
        users_served = int_or_none(connectivity_proportion, "nr_of_users")
        db.session.add(ConnectedProportion(
            nren_id=nren.id, nren=nren, year=year,
            user_category=user_type,
            coverage=coverage,
            number_connected=number_connected,
            market_share=market_share,
            users_served=users_served
        ))

    connectivity_levels = answers.get("connectivity_level", {})
    for user_type, connectivity_level in connectivity_levels.items():
        user_type = UserCategory[user_type]
        db.session.add(ConnectivityLevel(
            nren_id=nren.id, nren=nren, year=year,
            user_category=user_type,
            typical_speed=int_or_none(connectivity_level, "typical_speed"),
            highest_speed=int_or_none(connectivity_level, "highest_speed"),
            highest_speed_proportion=decimal_or_none(connectivity_level, "highest_speed_connection_percentage")
        ))

    traffic_carriers = answers.get("traffic_carriers", {})
    for user_type, traffic_carrier in traffic_carriers.items():
        user_type = UserCategory[user_type]
        traffic_carrier = traffic_carrier.get("carry_mechanism")
        if traffic_carrier:
            db.session.add(ConnectionCarrier(
                nren_id=nren.id, nren=nren, year=year,
                user_category=user_type,
                carry_mechanism=CarryMechanism[traffic_carrier]
            ))

    traffic_loads = answers.get("traffic_load", {})
    for user_type, traffic_load in traffic_loads.items():
        user_type = UserCategory[user_type]
        db.session.add(ConnectivityLoad(
            nren_id=nren.id, nren=nren, year=year,
            user_category=user_type,
            average_load_from_institutions=int_or_none(traffic_load, "average_from_institutions_to_network"),
            average_load_to_institutions=int_or_none(traffic_load, "average_to_institutions_from_network"),
            peak_load_from_institutions=int_or_none(traffic_load, "peak_from_institutions_to_network"),
            peak_load_to_institutions=int_or_none(traffic_load, "peak_to_institutions_from_network")
        ))

    traffic_growths = answers.get("traffic_growth", {})
    for user_type, traffic_growth in traffic_growths.items():
        user_type = UserCategory[user_type]
        db.session.add(ConnectivityGrowth(
            nren_id=nren.id, nren=nren, year=year,
            user_category=user_type,
            growth=decimal_or_zero(traffic_growth, "growth_rate")
        ))

    commercial_organizations = answers.get("commercial_organizations")
    if commercial_organizations:
        c1 = commercial_organizations.get("commercial_r_e", {}).get("connection")
        c2 = commercial_organizations.get("commercial_general", {}).get("connection")
        c3 = commercial_organizations.get("commercial_collaboration", {}).get("connection")
        c4 = commercial_organizations.get("commercial_service_provider", {}).get("connection")
        c5 = commercial_organizations.get("university_spin_off", {}).get("connection")
        db.session.add(CommercialConnectivity(
            nren_id=nren.id, nren=nren, year=year,
            commercial_r_and_e=CommercialConnectivityCoverage[c1] if c1 else None,
            commercial_general=CommercialConnectivityCoverage[c2] if c2 else None,
            commercial_collaboration=CommercialConnectivityCoverage[c3] if c3 else None,
            commercial_service_provider=CommercialConnectivityCoverage[c4] if c4 else None,
            university_spin_off=CommercialConnectivityCoverage[c5] if c5 else None,
        ))

    commercial_charging_levels = answers.get("commercial_charging_levels")
    if commercial_charging_levels:
        c1 = commercial_charging_levels.get("collaboration", {}).get("charging_level")
        c2 = commercial_charging_levels.get("services", {}).get("charging_level")
        c3 = commercial_charging_levels.get("peering", {}).get("charging_level")
        db.session.add(CommercialChargingLevel(
            nren_id=nren.id, nren=nren, year=year,
            collaboration=CommercialCharges[c1] if c1 else None,
            service_supplier=CommercialCharges[c2] if c2 else None,
            direct_peering=CommercialCharges[c3] if c3 else None,
        ))

    remote_campuses = answers.get("remote_campuses")
    if remote_campuses:
        remote_campuses = remote_campuses == "Yes"
        remote_campuses_specifics = answers.get("remote_campuses_specifics", [])
        remote_campuses_specifics = [
            {"country": i.get("country", ""), "local_r_and_e_connection": bool_or_none(i, "connected")}
            for i in remote_campuses_specifics
        ]
        db.session.add(RemoteCampuses(
            nren_id=nren.id, nren=nren, year=year,
            remote_campus_connectivity=remote_campuses,
            connections=remote_campuses_specifics
        ))

    dark_fibre_lease = answers.get("dark_fibre_lease")
    if dark_fibre_lease:
        dark_fibre_lease = dark_fibre_lease == "Yes"
        db.session.add(DarkFibreLease(
            nren_id=nren.id, nren=nren, year=year,
            iru_or_lease=dark_fibre_lease,
            fibre_length_in_country=int_or_none(answers, "dark_fibre_lease_kilometers_inside_country"),
            fibre_length_outside_country=int_or_none(answers, "dark_fibre_lease_kilometers_outside_country"),
            iru_duration=decimal_or_none(answers, "dark_fibre_lease_duration")
        ))

    dark_fibre_nren = answers.get("dark_fibre_nren")
    if dark_fibre_nren:
        dark_fibre_nren = dark_fibre_nren == "Yes"
        db.session.add(DarkFibreInstalled(
            nren_id=nren.id, nren=nren, year=year,
            installed=dark_fibre_nren,
            fibre_length_in_country=int_or_none(answers, "dark_fibre_nren_kilometers_inside_country")
        ))

    fibre_light = answers.get("fibre_light")
    if fibre_light:
        if fibre_light == "other":
            fibre_light = answers.get("fibre_light-Comment", "other")
        db.session.add(FibreLight(
            nren_id=nren.id, nren=nren, year=year,
            light_description=fibre_light
        ))

    network_map_urls = answers.get("network_map_urls", [])
    urls = [i.get("network_map_url", "") for i in network_map_urls if i.get("network_map_url", "") != ""]
    if urls:
        db.session.add(NetworkMapUrls(
            nren_id=nren.id, nren=nren, year=year,
            urls=urls
        ))

    monitoring_tools = answers.get("monitoring_tools", [])
    netflow_vendors = answers.get("netflow_vendors", "")
    if monitoring_tools or netflow_vendors:
        if "other" in monitoring_tools:
            monitoring_tools.remove("other")
            monitoring_tools.append(answers.get("monitoring_tools-Comment", "other"))
        db.session.add(MonitoringTools(
            nren_id=nren.id, nren=nren, year=year,
            tool_descriptions=monitoring_tools,
            netflow_processing_description=netflow_vendors
        ))

    passive_monitoring = answers.get("passive_monitoring")
    if passive_monitoring:
        passive_monitoring = passive_monitoring == "Yes"
        passive_monitoring_tech = answers.get("passive_monitoring_tech")
        db.session.add(PassiveMonitoring(
            nren_id=nren.id, nren=nren, year=year,
            monitoring=passive_monitoring,
            method=MonitoringMethod[passive_monitoring_tech] if passive_monitoring_tech else None
        ))

    traffic_statistics = answers.get("traffic_statistics")
    if traffic_statistics:
        traffic_statistics = traffic_statistics == "Yes"
        urls = answers.get("traffic_statistics_urls", [])
        urls = [i.get("traffic_statistics_url", "") for i in urls if i.get("traffic_statistics_url")]
        db.session.add(TrafficStatistics(
            nren_id=nren.id, nren=nren, year=year,
            traffic_statistics=traffic_statistics,
            urls=urls
        ))

    siem_soc_vendor = answers.get("siem_soc_vendor")
    if siem_soc_vendor:
        if "other" in siem_soc_vendor:
            siem_soc_vendor.remove("other")
            siem_soc_vendor.append(answers.get("siem_soc_vendor-Comment", "other"))
        db.session.add(SiemVendors(
            nren_id=nren.id, nren=nren, year=year,
            vendor_names=siem_soc_vendor
        ))

    certificate_service = answers.get("certificate_service")
    if certificate_service:
        if "other" in certificate_service:
            certificate_service.remove("other")
            certificate_service.append(answers.get("certificate_service-Comment", "other"))
        db.session.add(CertificateProviders(
            nren_id=nren.id, nren=nren, year=year,
            provider_names=certificate_service
        ))

    network_weather = answers.get("network_weather")
    if network_weather:
        network_weather = network_weather == "Yes"
        db.session.add(WeatherMap(
            nren_id=nren.id, nren=nren, year=year,
            weather_map=network_weather,
            url=answers.get("network_weather_url", "")
        ))

    pert_team = answers.get("pert_team")
    if pert_team:
        pert_team = YesNoPlanned[pert_team.lower()]
        db.session.add(PertTeam(
            nren_id=nren.id, nren=nren, year=year,
            pert_team=pert_team
        ))

    alienwave_services = answers.get("alienwave_services")
    alienwave_internal = answers.get("alienwave_internal")
    if alienwave_services or alienwave_internal:
        alienwave_services = YesNoPlanned[alienwave_services.lower()] if alienwave_services else None
        alienwave_internal = alienwave_internal == "Yes" if alienwave_internal else None
        db.session.add(AlienWave(
            nren_id=nren.id, nren=nren, year=year,
            alien_wave_third_pary=alienwave_services,
            nr_of_alien_wave_third_party_services=int_or_none(answers, "alienwave_services_number"),
            alien_wave_internal=alienwave_internal
        ))

    max_capacity = answers.get("max_capacity")
    typical_capacity = answers.get("typical_capacity")
    if max_capacity or typical_capacity:
        db.session.add(Capacity(
            nren_id=nren.id, nren=nren, year=year,
            largest_link_capacity=decimal_or_none(answers, "max_capacity"),
            typical_backbone_capacity=decimal_or_none(answers, "typical_capacity")
        ))

    external_connections = answers.get("external_connections")
    if external_connections:
        connections: List[ExternalConnection] = []
        for connection in external_connections:
            connections.append({
                'link_name': connection.get('link_name', ''),
                'capacity': connection.get('capacity'),
                'from_organization': connection.get('from_organization', ''),
                'to_organization': connection.get('to_organization', ''),
                'interconnection_method': connection.get('interconnection_method')
            })
        db.session.add(ExternalConnections(
            nren_id=nren.id, nren=nren, year=year,
            connections=connections
        ))

    non_r_and_e_peers = answers.get("non_r_and_e_peers")
    if non_r_and_e_peers:
        db.session.add(NonREPeers(
            nren_id=nren.id, nren=nren, year=year,
            nr_of_non_r_and_e_peers=int(non_r_and_e_peers)
        ))

    commodity_vs_r_e = answers.get("commodity_vs_r_e")
    if commodity_vs_r_e:
        db.session.add(TrafficRatio(
            nren_id=nren.id, nren=nren, year=year,
            r_and_e_percentage=decimal_or_zero(commodity_vs_r_e, "r_e"),
            commodity_percentage=decimal_or_zero(commodity_vs_r_e, "commodity"),
        ))

    operational_process_automation = answers.get("operational_process_automation")
    if operational_process_automation:
        db.session.add(OpsAutomation(
            nren_id=nren.id, nren=nren, year=year,
            ops_automation=YesNoPlanned[operational_process_automation.lower()],
            ops_automation_specifics=answers.get("operational_process_automation_tools", "")
        ))

    nfv = answers.get("nfv")
    if nfv:
        nfv_types = answers.get("nfv_types", [])
        if "other" in nfv_types:
            nfv_types.remove("other")
            nfv_types.append(answers.get("nfv_types-Comment", "other"))
        db.session.add(NetworkFunctionVirtualisation(
            nren_id=nren.id, nren=nren, year=year,
            nfv=YesNoPlanned[nfv.lower()],
            nfv_specifics=nfv_types
        ))

    network_automation = answers.get("network_automation")
    if network_automation:
        network_automation_tasks = answers.get("network_automation_tasks", [])
        db.session.add(NetworkAutomation(
            nren_id=nren.id, nren=nren, year=year,
            network_automation=YesNoPlanned[network_automation.lower()],
            network_automation_specifics=network_automation_tasks
        ))

    all_services = {}
    for question_name in ["services_collaboration", "services_hosting", "services_identity", "services_isp",
                          "services_multimedia", "services_network", "services_professional", "services_security"]:
        all_services.update(answers.get(question_name, {}))

    for service_key, answers in all_services.items():
        offered = answers.get("offered")
        if offered == ["yes"]:
            db.session.add(NRENService(
                nren_id=nren.id, nren=nren, year=year,
                service_key=service_key,
                product_name=answers.get("name", ""),
                additional_information=answers.get("additional_information", ""),
                official_description=answers.get("description", "")
            ))


def publish(year):
    responses = db.session.scalars(
        select(SurveyResponse).where(SurveyResponse.survey_year == year)
                              .where(SurveyResponse.status == ResponseStatus.completed)
    ).unique()

    question_mapping = {
        2023: _map_2023
    }

    mapping_function = question_mapping[year]

    for response in responses:
        mapping_function(response.nren, response.answers)
