"""
Utilities used by multiple route blueprints.
"""
import functools
import logging
from compendium_v2 import db
from compendium_v2.db.presentation_models import NREN, PreviewYear

from flask import Response, request
from flask_login import current_user  # type: ignore
from sqlalchemy import select

logger = logging.getLogger(__name__)


def require_accepts_json(f):
    """
    used as a route handler decorator to return an error
    unless the request allows responses with type "application/json"
    :param f: the function to be decorated
    :return: the decorated function
    """

    @functools.wraps(f)
    def decorated_function(*args, **kwargs):
        # TODO: use best_match to disallow */* ...?
        if not request.accept_mimetypes.accept_json:
            return Response(
                response='response will be json',
                status=406,
                mimetype='text/html')
        return f(*args, **kwargs)

    return decorated_function


def init_background_task():
    pass


def after_request(response):
    """
    Generic function to do additional logging of requests & responses.

    :param response:
    :return:
    """
    if response.status_code != 200:

        try:
            data = response.data.decode('utf-8')
        except Exception:
            # never expected to happen, but we don't want any failures here
            logging.exception('INTERNAL DECODING ERROR')
            data = 'decoding error (see logs)'

        logger.warning('"%s %s" "%s" %s' % (
            request.method,
            request.path,
            data,
            str(response.status_code)))
    return response


def get_data(table_class):
    select_statement = select(table_class).join(NREN).order_by(NREN.name.asc(), table_class.year.desc())
    is_admin = (not current_user.is_anonymous) and current_user.is_admin
    preview = is_admin and request.args.get('preview') is not None
    if not preview:
        select_statement = select_statement.where(table_class.year.not_in(select(PreviewYear.year)))
    return db.session.scalars(select_statement)
