import logging
from typing import Any

from flask import Blueprint, jsonify

from compendium_v2.db.presentation_models import ECProject
from compendium_v2.routes import common


routes = Blueprint('ec-projects', __name__)
logger = logging.getLogger(__name__)

EC_PROJECTS_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',

    'definitions': {
        'ec_project': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'nren_country': {'type': 'string'},
                'year': {'type': 'integer'},
                'project': {'type': 'string'}
            },
            'required': ['nren', 'nren_country', 'year', 'project'],
            'additionalProperties': False
        }
    },

    'type': 'array',
    'items': {'$ref': '#/definitions/ec_project'}
}


@routes.route('/', methods=['GET'])
@common.require_accepts_json
def ec_projects_view() -> Any:
    """
    handler for /api/ec-project requests
    returns EC projects for each NREN/year combination

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.ec_projects.EC_PROJECTS_RESPONSE_SCHEMA

    :return:
    """

    def _extract_project(entry: ECProject):
        return {
            'nren': entry.nren.name,
            'nren_country': entry.nren.country,
            'year': entry.year,
            'project': entry.project
        }

    result = sorted(
        [_extract_project(project) for project in common.get_data(ECProject)],
        key=lambda d: (d['nren'], d['year'], d['project'])
    )
    return jsonify(result)
