from typing import Any

from flask import Blueprint, jsonify

from compendium_v2.db.presentation_models import NRENService
from compendium_v2.routes import common

routes = Blueprint('nren-services', __name__)

NREN_SERVICES_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'definitions': {
        'nren_services': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'nren_country': {'type': 'string'},
                'year': {'type': 'integer'},
                'product_name': {'type': 'string'},
                'additional_information': {'type': 'string'},
                'official_description': {'type': 'string'},
                'service_name': {'type': 'string'},
                'service_category': {'type': 'string'},
                'service_description': {'type': 'string'}
            },
            'required': ['nren', 'nren_country', 'year', 'product_name', 'additional_information',
                         'official_description', 'service_name', 'service_category', 'service_description'],
            'additionalProperties': False
        }
    },
    'type': 'array',
    'items': {'$ref': '#/definitions/nren_services'}
}


@routes.route('/', methods=['GET'])
@common.require_accepts_json
def nren_service_view() -> Any:
    """
    handler for /api/nren-services/ requests
    Endpoint for getting the service matrix that shows which NREN uses which services.

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.nren_services.NREN_SERVICES_RESPONSE_SCHEMA

    :return:
    """

    def _extract_data(nren_service: NRENService) -> dict:
        return {
            'nren': nren_service.nren.name,
            'nren_country': nren_service.nren.country,
            'year': nren_service.year,
            'product_name': nren_service.product_name,
            'additional_information': nren_service.additional_information,
            'official_description': nren_service.official_description,
            'service_name': nren_service.service.name,
            'service_category': nren_service.service.category.value,
            'service_description': nren_service.service.description
        }

    entries = [_extract_data(entry) for entry in common.get_data(NRENService)]

    return jsonify(entries)
