import logging
from typing import Any

from flask import Blueprint, jsonify

from compendium_v2.db.presentation_models import ParentOrganization, SubOrganization
from compendium_v2.routes import common


routes = Blueprint('organization', __name__)
logger = logging.getLogger(__name__)

ORGANIZATION_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',

    'definitions': {
        'parent_organization': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'nren_country': {'type': 'string'},
                'year': {'type': 'integer'},
                'name': {'type': 'string'}
            },
            'required': ['nren', 'nren_country', 'year', 'name'],
            'additionalProperties': False
        },
        'sub_organization': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'nren_country': {'type': 'string'},
                'year': {'type': 'integer'},
                'name': {'type': 'string'},
                'role': {'type': 'string'},
            },
            'required': ['nren', 'nren_country', 'year', 'name', 'role'],
            'additionalProperties': False
        }
    },

    'type': 'array',
    'items': {
        'oneOf': [
            {'$ref': '#/definitions/parent_organization'},
            {'$ref': '#/definitions/sub_organization'}
        ]
    }
}


@routes.route('/parent', methods=['GET'])
@common.require_accepts_json
def parent_organization_view() -> Any:
    """
    handler for /api/organization/parent requests
    returns parent organizations for each NREN/year combination

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.organization.ORGANIZATION_RESPONSE_SCHEMA

    :return:
    """

    def _extract_parent(entry: ParentOrganization):
        return {
            'nren': entry.nren.name,
            'nren_country': entry.nren.country,
            'year': entry.year,
            'name': entry.organization
        }

    result = sorted(
        [_extract_parent(org) for org in common.get_data(ParentOrganization)],
        key=lambda d: (d['nren'], d['year'], d['name'])
    )

    return jsonify(result)


@routes.route('/sub', methods=['GET'])
@common.require_accepts_json
def sub_organization_view() -> Any:
    """
    handler for /api/organization/sub requests
    returns sub-organizations for each NREN/year combination

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.organization.ORGANIZATION_RESPONSE_SCHEMA

    :return:
    """

    def _extract_sub(entry: SubOrganization):
        return {
            'nren': entry.nren.name,
            'nren_country': entry.nren.country,
            'year': entry.year,
            'name': entry.organization,
            'role': entry.role
        }

    result = sorted(
        [_extract_sub(org) for org in common.get_data(SubOrganization)],
        key=lambda d: (d['nren'], d['year'], d['name'])
    )
    return jsonify(result)
