import React, { useState, useEffect, useRef } from "react";
import Accordion from 'react-bootstrap/Accordion';
import Button from 'react-bootstrap/Button';
import Table from 'react-bootstrap/Table';
import toast, { Toaster } from "react-hot-toast";
import { useNavigate } from 'react-router-dom';
import { SurveyStatus } from "./Schema";
import { fetchSurveys } from "./api/survey";
import { Survey } from "./api/types";
import { Spinner } from "react-bootstrap";


function ApiCallButton({ text, helpText, onClick, enabled }) {
    const [buttonLoading, setButtonLoading] = useState(false);

    const handleClick = async () => {
        if (buttonLoading) {
            return;
        }
        setButtonLoading(true);
        try {
            await onClick();
        } finally {
            setButtonLoading(false);
        }
    };

    return (
        <Button onClick={handleClick} disabled={!enabled} style={{ pointerEvents: 'auto', marginLeft: '.5rem' }} title={helpText}>
            {buttonLoading && <Spinner as="span" animation="border" size="sm" role="status" aria-hidden="true"/>}
            {text}
        </Button>
    );
}


function SurveyManagementComponent() {
    const [surveys, setSurveys] = useState<Survey[]>([]);
    const surveyStatusUpdating = useRef(false);

    useEffect(() => {
        fetchSurveys().then((surveyList) => {
            setSurveys(surveyList);
        });
    }, []);

    async function apiCall(url, failureToast, successToast) {
        try {
            const response = await fetch(url, { method: 'POST' });
            const json = await response.json();
            if (!response.ok) {
                toast(failureToast + json['message']);
            } else {
                toast(successToast);
                fetchSurveys().then((surveyList) => {
                    setSurveys(surveyList);
                });
            }
        } catch(e) {
            toast(failureToast + (e as Error).message);
        }
    }

    async function newSurvey() {
        await apiCall('/api/survey/new', "Failed creating new survey: ", "Created new survey");
    }

    async function postSurveyStatus(year, status) {
        if (surveyStatusUpdating.current) {
            toast("Wait for status update to be finished...");
            return;
        }
        surveyStatusUpdating.current = true;
        await apiCall(
            '/api/survey/' + status + '/' + year,
            "Error while updating " + year + " survey status to " + status + ": ",
            year + " survey status updated to " + status
        );
        surveyStatusUpdating.current = false;
    }

    async function removeLock(year, nrenName) {
        await apiCall(
            '/api/response/unlock/' + year + '/' + nrenName,
            "Error while unlocking " + nrenName + " " + year + " survey response: ",
            nrenName + " " + year + " survey response unlocked"
        );
    }

    const newSurveyAllowed = surveys.length > 0 && surveys.every(s => s.status == SurveyStatus.published);

    const navigate = useNavigate();

    const previewLink = window.location.origin + "/data?preview";

    return (
        <div>
            <Toaster />
            <Button onClick={newSurvey} disabled={!newSurveyAllowed} style={{ pointerEvents: 'auto' }}
                title="Create a new survey for the next year. Only possible if all current surveys are published.">
                start new survey
            </Button>
            <Accordion>
                {surveys.map(survey => (
                    <Accordion.Item eventKey={survey.year.toString()} key={survey.year}>
                        <Accordion.Header>{survey.year} - {survey.status}</Accordion.Header>
                        <Accordion.Body>
                            <div style={{ marginLeft: '.5rem' }}>
                                <Button style={{ marginLeft: '.5rem' }} onClick={() => navigate(`/survey/admin/inspect/${survey.year}`)}
                                    title="Open the survey for inspection with all questions visible and any visibleIf logic added to the title.">
                                    Inspect Survey
                                </Button>
                                <Button style={{ marginLeft: '.5rem' }} onClick={() => navigate(`/survey/admin/try/${survey.year}`)}
                                    title="Open the survey exactly as the nrens will see it, but without any nren data.">
                                    Try Survey
                                </Button>
                                <ApiCallButton text="Mark as open"
                                    helpText="Allow the NRENs to respond to this survey. Only 1 survey may be open at a time, and (pre)-published surveys cannot be opened anymore."
                                    enabled={survey.status == SurveyStatus.closed}
                                    onClick={() => postSurveyStatus(survey.year, 'open')}
                                />
                                <ApiCallButton text="Mark as closed"
                                    helpText="Do not allow the NRENs to respond to this survey anymore. Only surveys with status open can be closed."
                                    enabled={survey.status == SurveyStatus.open}
                                    onClick={() => postSurveyStatus(survey.year, 'close')}
                                />
                                <ApiCallButton text="Preview results"
                                    helpText="Publish all completed survey responses to the compendium website for preview by admins. This is only possible if the survey is closed or previewed already."
                                    enabled={survey.status == SurveyStatus.closed || survey.status == SurveyStatus.preview}
                                    onClick={() => postSurveyStatus(survey.year, 'preview')}
                                />
                                <ApiCallButton text="Publish results"
                                    helpText="Publish or re-publish all completed survey responses to the compendium website. This is only possible if the survey is in preview or published already."
                                    enabled={survey.status == SurveyStatus.preview || survey.status == SurveyStatus.published}
                                    onClick={() => postSurveyStatus(survey.year, 'publish')}
                                />
                                {survey.status == SurveyStatus.preview && <span>&nbsp; Preview link: <a href={previewLink}>{previewLink}</a></span>}
                            </div>
                            <Table>
                                <tbody>
                                    {survey.responses.map(response => (
                                        <tr key={response.nren}>
                                            <td>{response.nren}</td>
                                            <td>{response.status}</td>
                                            <td>{response.lock_description}</td>
                                            <td>
                                                <Button onClick={() => navigate(`/survey/response/${survey.year}/${response.nren}`)} style={{ pointerEvents: 'auto' }}
                                                    title="Open the responses of the NREN.">
                                                    open
                                                </Button>
                                                <Button onClick={() => removeLock(survey.year, response.nren)} disabled={response.lock_description == ''} style={{ pointerEvents: 'auto' }}
                                                    title="Remove the lock from the survey so that another person can open the survey for editing. WARNING: The person that currently has the lock will not be abe to save their changes anymore once someone else starts editing!">
                                                    remove lock
                                                </Button>
                                            </td>
                                        </tr>
                                    ))}

                                </tbody>
                            </Table>

                        </Accordion.Body>
                    </Accordion.Item>
                ))}
            </Accordion>
        </div>
    );

}

export default SurveyManagementComponent;
