import React, { useState, useEffect, useContext } from "react";
import { Container, Row, Table } from "react-bootstrap";
import { userContext } from "./providers/UserProvider";
import { User, Nren } from "./Schema";


async function fetchUsers(): Promise<User[]> {
    try {
        const response = await fetch('/api/user/list');
        const userList = await response.json();
        return userList
    } catch (error) {
        console.log('handle this better..');
        return [];
    }
}

async function fetchNrens(): Promise<Nren[]> {
    try {
        const response = await fetch('/api/nren/list');
        const userList = await response.json();
        return userList
    } catch (error) {
        console.log('handle this better..');
        return [];
    }
}

const updateUser = async (id, changes) => {

    const body = {
        id: id,
        ...changes
    }
    const requestOptions = {
        method: 'PUT',
        headers: {
            'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
    };

    const response = await fetch(`/api/user/${id}`, requestOptions)
    const data = await response.json()
    if (!response.ok) {
        throw new Error(data.message);
    }

    return data.user;
};

const defaultSortFunction = (a, b) => {
    if (a.permissions.active && !b.permissions.active) {
        return -1;
    } else if (!a.permissions.active && b.permissions.active) {
        return 1;
    } else if (a.permissions.active && b.permissions.active) {
        if (a.role === 'admin' && b.role !== 'admin') {
            return 1;
        } else if (a.role !== 'admin' && b.role === 'admin') {
            return -1;
        } else {
            return a.name.localeCompare(b.name)
        }
    } else {
        return a.name.localeCompare(b.name)
    }
}


function UserManagementComponent() {
    const [users, setUsers] = useState<User[]>([]);
    const [nrens, setNrens] = useState<Nren[]>([]);
    const { user: loggedInUser } = useContext(userContext);
    const [sortColumn, setSortColumn] = useState({ idx: -1, asc: true });
    const [sortedUsers, setSortedUsers] = useState<User[]>([]);

    useEffect(() => {
        fetchUsers().then((userList) => {
            setUsers(userList);
            setSortedUsers(userList.sort(defaultSortFunction))
        });

        fetchNrens().then((nrenList) => {
            setNrens(nrenList.sort((a, b) => a.name.localeCompare(b.name)))
        })
    }, []);

    const handleInputChange = (event: React.ChangeEvent<HTMLInputElement | HTMLSelectElement>, user: User) => {
        const index = users.findIndex((u) => u.id === user.id);
        const updatedUsers = [...users];

        const { name } = event.target;

        const update = {};

        if (name === 'active') {
            update[name] = (event.target as HTMLInputElement).checked
        } else {
            update[name] = event.target.value;
        }

        updateUser(user.id, update).then((user) => {
            updatedUsers[index] = user;
            setUsers(updatedUsers);
        }).catch((error) => {
            alert(error.message);
        });
    };

    const findNren = (value) => {
        return nrens.find((nren) => {
            return nren.id == value || nren.name == value
        })?.id
    }

    const setSort = (index) => {
        let func;
        if (index === sortColumn.idx || ((index === 5 || index === 0) && sortColumn.idx === -1)) {
            // reverse sort
            if (index === 5 || index === 0) {
                // hack not to show the sort icon on column 0 and 5
                index = -1
            }
            setSortColumn({ idx: index, asc: !sortColumn.asc })
            setSortedUsers([...sortedUsers.reverse()])
            return
        }
        
        if (index === 0) {
            // user ID is a UUID which is meaningless to sort by, so use default sort function
            func = defaultSortFunction
            setSortColumn({ idx: -1, asc: true })
        } else if (index === 1) {
            // sort by active
            func = (a, b) => {
                if (a.permissions.active && !b.permissions.active) {
                    return -1;
                } else if (!a.permissions.active && b.permissions.active) {
                    return 1;
                } else {
                    return 0;
                }
            }
            setSortColumn({ idx: index, asc: true })
        } else if (index === 2) {
            // sort by role
            func = (a, b) => {
                return a.role.localeCompare(b.role)
            }
            setSortColumn({ idx: index, asc: true })
        } else if (index === 3) {
            // sort by email
            func = (a, b) => {
                return a.email.localeCompare(b.email)
            }
            setSortColumn({ idx: index, asc: true })
        } else if (index === 4) {
            // sort by name
            func = (a, b) => {
                return a.name.localeCompare(b.name)
            }
            setSortColumn({ idx: index, asc: true })
        } else if (index === 5) {
            // use the default sort function, OIDC sub has no meaning when sorting
            func = defaultSortFunction
            setSortColumn({ idx: -1, asc: true })
        } else if (index === 6) {
            // sort by NREN
            func = (a, b) => {
                if (a.nrens.length === 0 && b.nrens.length === 0) {
                    return 0;
                } else if (a.nrens.length === 0) {
                    return -1;
                } else if (b.nrens.length === 0) {
                    return 1;
                } else {
                    return a.nrens[0].localeCompare(b.nrens[0])
                }
            }
            setSortColumn({ idx: index, asc: true })
        } else {
            func = defaultSortFunction
            setSortColumn({ idx: index, asc: true })
        }
        setSortedUsers(users.sort(func))
    }

    // build the aria-sort attribute for each column with spreadable objects..
    // this is apparently the easiest way to do it conditionally
    const ariaSort = {}

    for (let i = 0; i <= 6; i++) {
        ariaSort[i] = sortColumn.idx === i ? ({ 'aria-sort': sortColumn.asc ? 'ascending' : 'descending' }) : null
    }

    return (
        <Container style={{ maxWidth: '90vw', }}>
            <Row>
                <h1> User Management Page</h1>
                <Table>
                    <thead>
                        <tr>
                            <th {...ariaSort[0]} onClick={() => setSort(0)} className='pt-3 sortable' style={{ border: "1px solid #ddd" }}>Id</th>
                            <th {...ariaSort[1]} onClick={() => setSort(1)} className='pt-3 sortable' style={{ border: "1px solid #ddd" }}>Active</th>
                            <th {...ariaSort[2]} onClick={() => setSort(2)} className='pt-3 sortable' style={{ border: "1px solid #ddd" }}>Role</th>
                            <th {...ariaSort[3]} onClick={() => setSort(3)} className='pt-3 sortable' style={{ border: "1px solid #ddd" }}>Email</th>
                            <th {...ariaSort[4]} onClick={() => setSort(4)} className='pt-3 sortable' style={{ border: "1px solid #ddd" }}>Full Name</th>
                            <th {...ariaSort[5]} onClick={() => setSort(5)} className='pt-3 sortable' style={{ border: "1px solid #ddd" }}>OIDC Sub</th>
                            <th {...ariaSort[6]} onClick={() => setSort(6)} className='pt-3 sortable' style={{ border: "1px solid #ddd" }}>NREN</th>
                        </tr>
                    </thead>
                    <tbody>
                        {sortedUsers.map(user => (
                            <tr key={user.id}>
                                <td style={{ border: "1px dotted #ddd" }}>{user.id}</td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.id == loggedInUser.id ? 'Active' : <input
                                        type="checkbox"
                                        name="active"
                                        checked={user.permissions.active}
                                        onChange={(event) => handleInputChange(event, user)}
                                    />}

                                </td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.id == loggedInUser.id ? (user.role.charAt(0).toUpperCase()
                                        + user.role.slice(1)) : <select
                                            name="role"
                                            defaultValue={user.role}
                                            onChange={(event) => handleInputChange(event, user)}>
                                        <option value="admin">Admin</option>
                                        <option value="user">User</option>
                                        <option value="observer">Observer</option>
                                    </select>}
                                </td>

                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.email}
                                </td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.name}
                                </td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.oidc_sub}
                                </td>
                                <td style={{ border: '1px dotted #ddd' }}>
                                    <select
                                        name="nren"
                                        multiple={false}
                                        defaultValue={user.nrens.length > 0 ? findNren(user.nrens[0]) : undefined}
                                        onChange={(event) => handleInputChange(event, user)}>
                                        <option>
                                            Select NREN
                                        </option>
                                        {nrens.map((nren) => (
                                            <option key={nren.id} value={nren.id}>
                                                {nren.name}
                                            </option>
                                        ))}
                                    </select>
                                </td>
                            </tr>
                        ))}
                    </tbody>

                </Table>
            </Row>

        </Container>
    );

}

export default UserManagementComponent;
